import re
from email.mime.base import MIMEBase
from logging import getLogger
from pathlib import Path
from typing import Any, Mapping, Optional, Sequence, Union

import premailer
from antimarkdown import to_markdown
from anymail.message import AnymailMessage, AnymailStatus
from django.conf import settings
from django.template.loader import get_template
from jinja2.utils import concat

logger = getLogger(__name__)


class Email:
    """
    Create an email based on a template.

    .. note:: Requires the :ref:`dynamic extra <index:Dynamic Sites>` and a valid :doc:`Anymail
        <django-anymail:index>` configuration.

    .. tip:: We recommend using the standard :ref:`dynamic site settings <settings:Dynamic Site
        Settings>` module, which configures the :doc:`Amazon SES Anymail ESP
        <django-anymail:esps/amazon_ses>` with the standard :ref:`Stormware AWS authentication
        <stormware:auth:Amazon Web Services>` process.

    Args:
        template: The template to use.
        context: The context to use for template rendering.
        prefix: The prefix to use for the subject.
            Defaults to the value of the ``EMAIL_SUBJECT_PREFIX`` setting.
            Note that a space is automatically appended to the provided value.

    Template
    ========
    .. tip:: We recommend extending the :ref:`standard email base template <templates:Email>`.

    Blocks
    ------
    .. py:data:: subject
        :noindex:

        The email's subject *(required)*.

    .. py:data:: text
        :noindex:

        The plain text version of the email *(optional)*. Defaults to the Markdown version of the
        rendered HTML content.

    Functions
    ---------
    .. py:function:: image(path: typing.Union[str, ~pathlib.Path]) -> str
        :noindex:

        Add an inline image from a local path to the email and return the image source string.

    Methods & Properties
    ====================

    """
    def __init__(
        self,
        template: str,
        context: Optional[Mapping[str, Any]] = None,
        prefix: Optional[str] = None,
    ):
        # Loading template
        loaded_template = get_template(template)
        self._template = loaded_template.template  # type: ignore[attr-defined]
        self._template_context = self._template.new_context(context or {})

        # Creating email instance
        if (prefix := prefix if prefix is not None else settings.EMAIL_SUBJECT_PREFIX.strip()):
            prefix += ' '
        self._email = AnymailMessage(subject=f'{prefix}{self._get_block("subject")}')

        # Rendering html
        from django.contrib.sites.models import Site  # pylint: disable=import-outside-toplevel

        html = loaded_template.render({**self._template_context, 'image': self._attach_image})
        base_url = (
            'http://127.0.0.1:8000' if settings.DEBUG
            else f'https://{Site.objects.get_current().domain}'
        )
        html = premailer.transform(html, base_url=base_url, allow_network=False)
        self._email.attach_alternative(html, mimetype='text/html')

        # Rendering text
        if (text := self._get_block('text', required=False)) is None:
            text = re.sub('<script[^<]*</script>', '', html)  # remove script elements
            text = re.sub('\n +', '\n', text)  # remove indentation
            text = text.replace('<div', '<p')  # fix div start tag
            text = text.replace('</div>', '</p>')  # fix div end tag
            text = to_markdown(text) or ''
            text = re.sub(r'([=\-#])\n+', r'\1\n', text)  # fix heading
            text = re.sub(r'^\* +', r'- ', text, flags=re.MULTILINE)  # fix unordered list
            text = re.sub(r'^([0-9]+\.) +', r'\1 ', text, flags=re.MULTILINE)  # fix ordered list
            text = text.replace('\n\n\n', '\n\n')  # fix extra newlines
            text = re.sub(r'^---$', '---\n', text, flags=re.MULTILINE)  # fix horizontal rule
        self._email.body = text.strip() + '\n'

    def _get_block(self, name: str, required: bool = True) -> Optional[str]:
        block = self._template.blocks.get(name)
        if required and not block:
            raise RuntimeError(f'The "{name}" block must be defined in the tempate')
        return concat(block(self._template_context)).strip() if block else None

    def _attach_image(self, path: Union[str, Path], subtype: Optional[str] = None) -> str:
        return f'cid:{self._email.attach_inline_image_file(path=path, subtype=subtype)}'

    def attach(
        self, filename: Union[str, MIMEBase],
        content: Optional[Any] = None,
        mimetype: Optional[str] = None,
    ) -> None:
        """
        Create a new file attachment and add it to the message.

        See ``attach()`` in :class:`django.core.mail.EmailMessage`.
        """
        self._email.attach(filename=filename, content=content, mimetype=mimetype)

    def attach_file(self, path: Union[str, Path], mimetype: Optional[str] = None) -> None:
        """
        Create a new attachment using a file from the filesystem.

        See ``attach_file()`` in :class:`django.core.mail.EmailMessage`.
        """
        self._email.attach_file(path=path, mimetype=mimetype)

    def send(  # pylint: disable=too-many-arguments
        self,
        sender: str,
        to: Sequence[str],
        cc: Optional[Sequence[str]] = None,
        bcc: Optional[Sequence[str]] = None,
        reply_to: Optional[Sequence[str]] = None,
    ) -> int:
        """
        Send the email.

        Args:
            sender: The sender's address.
            to: A list of recipients.
            cc: A list of carbon copy recipients.
            bcc: A list of blind carbon copy recipients.
            reply_to: A list of addresses for replies.

        """
        logger.info('Sending email')
        self._email.from_email = sender
        self._email.to = to
        self._email.cc = cc or []
        self._email.bcc = bcc or []
        self._email.reply_to = reply_to or []
        response = self._email.send()
        info = (
            f' (status: {self.status.status}, ID {self.status.message_id})'
            if self.status.message_id is not None else ''
        )
        logger.info(f'Email sent{info}')
        return response  # type: ignore[no-any-return]

    @property
    def status(self) -> AnymailStatus:
        """
        Return the status of the email.
        """
        return self._email.anymail_status



"""
Start the Django development web server.
"""
import argparse
import sys
from os import environ
from typing import Sequence

from django.core.management import execute_from_command_line
from logikal_utils.project import tool_config

from django_logikal.env import set_option


def main(args: Sequence[str] = tuple(sys.argv[1:])) -> int:
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument('-d', '--debug', action='store_true', help='use debug logging level')
    parser.add_argument('-t', '--toolbar', action='store_true', help='enable the debug toolbar')
    parser.add_argument('--cloud-logging', action='store_true', help='use cloud logging')
    parser.add_argument('--send-emails', action='store_true', help='send out emails')
    util_args, manage_args = parser.parse_known_args(args)
    if util_args.debug:
        set_option('log_level', 'DEBUG')
    if util_args.toolbar:
        set_option('toolbar')
    if util_args.cloud_logging:
        set_option('cloud_logging')
    if util_args.send_emails:
        set_option('send_emails')

    sys.path.insert(0, '.')
    set_option('dev_run')
    environ.setdefault(
        'DJANGO_SETTINGS_MODULE',
        tool_config('django_logikal')['DJANGO_SETTINGS_MODULE'],
    )
    execute_from_command_line(['manage', 'runserver'] + manage_args)
    return 0



"""
Generate static files.

.. note:: Requires the :ref:`static extra <index:Static Sites>`.
"""
from typing import Any, Mapping

from django.core.management import call_command
from django.core.management.base import BaseCommand, CommandParser


class Command(BaseCommand):
    help = ' '.join(__doc__.splitlines()[0:2])

    def add_arguments(self, parser: CommandParser) -> None:
        parser.add_argument('--no-input', action='store_true', help='Do not prompt for input.')
        parser.add_argument('output_dir', nargs='?', help='The output folder to use.')

    def handle(self, *_args: Any, **options: Mapping[str, Any]) -> None:
        no_input = options.pop('no_input', False)
        output_dir = options.pop('output_dir', None)
        call_command('collectstatic', clear=True, no_input=not no_input, **options)
        call_command('distill-local', force=no_input, output_dir=output_dir, **options)



from django_logikal.settings.common.testing import CommonTestingSettings
from django_logikal.settings.static_site.dev import DevSettings


class TestingSettings(CommonTestingSettings, DevSettings):
    """
    Standard settings for testing static sites.
    """



import os
from pathlib import Path

from django_logikal.settings import Settings
from django_logikal.settings.common.base import CommonBaseSettings


class BaseSettings(CommonBaseSettings):
    MIDDLEWARE = [
        'django.middleware.locale.LocaleMiddleware',  # required for localization
        'django.middleware.common.CommonMiddleware',  # performs URL rewriting and sets headers
        'django.contrib.sites.middleware.CurrentSiteMiddleware',  # adds site attribute to requests
    ]

    # Static site generation
    DISTILL_DIR = Path(os.getcwd()) / 'generated'

    @staticmethod
    def apply(settings: Settings) -> None:
        settings['INSTALLED_APPS'] += ['django_distill']



import os

from logikal_utils.docker import Service

from django_logikal.env import get_option, option_is_set
from django_logikal.logging import logging_config
from django_logikal.settings import Settings, SettingsUpdate


class CommonDevSettings(SettingsUpdate):
    # Logging
    LOGGING = logging_config(
        log_level=get_option('log_level', 'INFO'),
        console=True,
        cloud=option_is_set('cloud_logging'),
    )

    # Security
    SECRET_KEY = 'dev'  # nosec: only used for development

    # Core settings
    DEBUG = option_is_set('dev_run')
    DATABASES = {
        'default': {
            'ENGINE': 'django.db.backends.postgresql',
            'HOST': os.getenv('DJANGO_DATABASE_HOST', '127.0.0.1'),
            'PORT': os.getenv(
                'DJANGO_DATABASE_PORT',
                Service('postgres').container_port('5432/tcp'),
            ),
            'NAME': os.getenv('DJANGO_DATABASE_NAME', 'dev'),
            'USER': os.getenv('DJANGO_DATABASE_USER', 'dev'),
            'PASSWORD': os.getenv('DJANGO_DATABASE_USER', 'dev'),
        },
    }

    # Migration linter
    MIGRATION_LINTER_OPTIONS = {'exclude_apps': ['sites', 'robots']}
    MIGRATION_LINTER_OVERRIDE_MAKEMIGRATIONS = True

    @staticmethod
    def apply(settings: Settings) -> None:
        # HTML validation
        settings['MIDDLEWARE'] += ['django_logikal.validation.ValidationMiddleware']

        # Migration linter
        settings['INSTALLED_APPS'] += ['django_migration_linter']

        # Debug toolbar
        if option_is_set('toolbar'):  # pragma: no cover, tested in subprocess
            settings['MIDDLEWARE'].insert(0, 'debug_toolbar.middleware.DebugToolbarMiddleware')
            settings['CSP_DEFAULT_SRC'] += " 'nonce-debug-toolbar'"
            settings['INSTALLED_APPS'] += ['debug_toolbar', 'template_profiler_panel']
            settings['INTERNAL_IPS'] = ['127.0.0.1']
            settings['DEBUG_TOOLBAR_PANELS'] = [
                'debug_toolbar.panels.history.HistoryPanel',
                'debug_toolbar.panels.timer.TimerPanel',
                'debug_toolbar.panels.settings.SettingsPanel',
                'debug_toolbar.panels.headers.HeadersPanel',
                'debug_toolbar.panels.request.RequestPanel',
                'debug_toolbar.panels.sql.SQLPanel',
                'template_profiler_panel.panels.template.TemplateProfilerPanel',
                'debug_toolbar.panels.staticfiles.StaticFilesPanel',
                'debug_toolbar.panels.cache.CachePanel',
                'debug_toolbar.panels.signals.SignalsPanel',
                'debug_toolbar.panels.redirects.RedirectsPanel',
                'debug_toolbar.panels.profiling.ProfilingPanel',
            ]
            settings['DEBUG_TOOLBAR_CONFIG'] = {
                'SHOW_COLLAPSED': True,
                'TOOLBAR_LANGUAGE': 'en-us',
                'SQL_WARNING_THRESHOLD': 100,  # milliseconds
                'DISABLE_PANELS': {
                    'debug_toolbar.panels.signals.SignalsPanel',
                    'debug_toolbar.panels.redirects.RedirectsPanel',
                    'debug_toolbar.panels.profiling.ProfilingPanel',
                },
            }
            if 'EMAIL_BACKEND' in settings:
                settings['INSTALLED_APPS'] += ['mail_panel']
                settings['EMAIL_BACKEND'] = 'mail_panel.backend.MailToolbarBackend'
                settings['DEBUG_TOOLBAR_PANELS'].insert(0, 'mail_panel.panels.MailToolbarPanel')



"""
Synchronize the local database.

Clear the public schema, apply migrations and insert local application data.

.. note:: Requires :ref:`pytest-logikal[django] <pytest-logikal:index:django>` to be installed.
.. note:: Local data must be specified in the ``local_data.py`` submodule of a given application in
    classes inheriting from :class:`~django_logikal.local_data.LocalData`.

.. autoclass:: django_logikal.local_data.LocalData
    :undoc-members:

.. tip:: We recommend using :doc:`factory_boy <factory-boy:index>`'s
    :class:`~factory.django.DjangoModelFactory` for generating synthetic data. By registering the
    factories in your tests (via `pytest-factoryboy
    <https://pytest-factoryboy.readthedocs.io/en/stable/>`_) you can use the same synthetic data
    for local development and testing.

"""
from contextlib import suppress
from importlib import import_module
from typing import Any

from django.apps import apps
from django.core.management import call_command
from django.core.management.base import BaseCommand, CommandError, CommandParser
from django.db import connections
from factory import random as factory_random

from django_logikal.local_data import LocalData


class Command(BaseCommand):
    help = ' '.join(__doc__.splitlines()[0:4])

    def add_arguments(self, parser: CommandParser) -> None:
        parser.add_argument('--no-input', action='store_true', help='Do not prompt for input.')

    def handle(self, *_args: Any, **options: Any) -> None:
        connection = connections['default']
        database = connection.settings_dict
        if database['HOST'] != '127.0.0.1':
            raise CommandError('Only the local database can be synchronized')

        database_url = f'{database["HOST"]}:{database["PORT"]}/{database["NAME"]}'
        self.stdout.write(f'Synchronizing database {self.style.ERROR(database_url)}')
        warning = '\nThis will DELETE ALL CURRENT DATA in the public schema!'
        self.stdout.write(self.style.WARNING(warning))
        self.stdout.write('\nAre you sure you want to do this?')
        prompt = 'Type \'yes\' to continue, or \'no\' to cancel: '
        if not options.get('no_input') and input(prompt) != 'yes':
            raise CommandError('Cancelled')

        # Clear schema
        self.stdout.write(self.style.MIGRATE_HEADING('\nClearing public schema:'))
        with connection.cursor() as cursor:
            self.stdout.write('  Dropping schema...', ending='')
            cursor.execute('DROP SCHEMA public CASCADE')
            self.stdout.write(self.style.SUCCESS(' OK'))
            self.stdout.write('  Creating schema...', ending='')
            cursor.execute('CREATE SCHEMA public')
            self.stdout.write(self.style.SUCCESS(' OK'))

        # Run migrations
        self.stdout.write('')
        call_command('migrate', **options)

        # Insert local data
        from pytest_logikal import django  # pylint: disable=import-outside-toplevel

        factory_random.reseed_random(django.DEFAULT_RANDOM_SEED)  # for deterministic data
        for app in apps.get_app_configs():
            with suppress(ImportError):
                if (module := getattr(app, 'module', None)):
                    import_module(f'{module.__name__}.local_data')  # register LocalData subclasses
        if LocalData.__subclasses__():
            self.stdout.write(self.style.MIGRATE_HEADING('\nInserting local data:'))
        for local_data in LocalData.__subclasses__():
            if issubclass(local_data, LocalData):
                module = local_data.__module__
                name = local_data.__name__
                self.stdout.write(f'  Inserting {module}.{name}...', ending='')
                local_data.insert()
                self.stdout.write(self.style.SUCCESS(' OK'))

        self.stdout.write(self.style.SUCCESS('\nDatabase successfully synchronized'))



import logging
import os
from pathlib import Path

from django.utils.translation import gettext_lazy

from django_logikal.settings import Settings, SettingsUpdate


class CommonBaseSettings(SettingsUpdate):
    ALLOWED_HOSTS = ['.localhost', '127.0.0.1', '[::1]']  # variants of localhost

    # Note: we cannot use UUID fields as a default (https://code.djangoproject.com/ticket/32577)
    DEFAULT_AUTO_FIELD = 'django.db.models.BigAutoField'

    INSTALLED_APPS = [
        'django.contrib.auth',  # required by sites, Django admin and the auth framework
        'django.contrib.contenttypes',  # required by sites, Django admin and the auth framework
        'django.contrib.sites',  # needed for the sitemap and the robots application
        'django.contrib.sitemaps',
        'django.contrib.staticfiles',
        'robots',
        'django_logikal',
    ]
    STORAGES = {
        'default': {'BACKEND': 'django.core.files.storage.FileSystemStorage'},
        'staticfiles': {'BACKEND': 'django_logikal.static.ManifestStaticFilesStorage'},
    }
    TEMPLATES = [{
        'BACKEND': 'django_logikal.templates.jinja.JinjaTemplates',
        'APP_DIRS': True,
    }, {
        # Needed for the bibliography module (and Django admin)
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'APP_DIRS': True,
        'OPTIONS': {
            'context_processors': [
                'django.template.context_processors.request',  # required by Django admin
                'django.contrib.auth.context_processors.auth',  # required by Django admin
                'django.contrib.messages.context_processors.messages',  # required by Django admin
            ],
        },
    }]
    SITE_ID = 1  # ID of the site created by default
    ROBOTS_SITEMAP_VIEW_NAME = 'sitemap'

    # Static files
    STATIC_ROOT = Path(os.getcwd()) / 'static'
    STATIC_URL = '/static/'

    # Internationalization
    USE_I18N = True
    LANGUAGES = [('en-us', gettext_lazy('English (United States)'))]
    LANGUAGE_CODE = 'en-us'
    USE_TZ = True
    TIME_ZONE = 'Europe/Zurich'

    @staticmethod
    def apply(settings: Settings) -> None:
        # Logging
        logging.captureWarnings(capture=True)
        logging.getLogger('blib2to3').setLevel(logging.WARNING)  # Black migration formatting



from django_logikal.settings import Settings
from django_logikal.settings.common.testing import CommonTestingSettings
from django_logikal.settings.dynamic_site.dev import DevSettings


class TestingSettings(CommonTestingSettings, DevSettings):
    """
    Standard settings for testing dynamic sites.
    """
    @staticmethod
    def apply(settings: Settings) -> None:
        settings['MIDDLEWARE'] = [
            middleware for middleware in settings['MIDDLEWARE']
            if 'WhiteNoiseMiddleware' not in middleware
        ]



import json

from logikal_utils.project import PYPROJECT
from stormware.google.secrets import SecretManager

from django_logikal.logging import logging_config
from django_logikal.settings import Settings, SettingsUpdate


class CommonProductionSettings(SettingsUpdate):
    # Core settings
    DEBUG = False
    CONN_MAX_AGE = 1 * 60 * 60  # 1 hour
    CONN_HEALTH_CHECKS = True

    # Security
    SECRET_KEY_PATH = f'{PYPROJECT["project"]["name"]}-website-secret-key'  # nosec: this is a path
    DATABASE_SECRETS_PATH = f'{PYPROJECT["project"]["name"]}-website-database-secrets'

    CSRF_COOKIE_SECURE = True
    LANGUAGE_COOKIE_SECURE = True
    SECURE_SSL_REDIRECT = True
    SESSION_COOKIE_SECURE = True

    # Email
    EMAIL_BACKEND = 'anymail.backends.amazon_ses.EmailBackend'

    @staticmethod
    def apply(settings: Settings) -> None:
        settings['LOGGING'] = logging_config(console=False, cloud=True)

        with SecretManager() as secrets:
            settings['SECRET_KEY'] = secrets[settings['SECRET_KEY_PATH']]
            database_secrets = json.loads(secrets[settings['DATABASE_SECRETS_PATH']])

        settings['DATABASES'] = {
            'default': {
                'ENGINE': 'django.db.backends.postgresql',
                'HOST': database_secrets['hostname'],
                'PORT': database_secrets['port'],
                'NAME': database_secrets['database'],
                'USER': database_secrets['username'],
                'PASSWORD': database_secrets['password'],
            },
        }



# pylint: disable=import-outside-toplevel
from django.db.migrations.writer import MigrationWriter


class FormattedMigrationWriter(MigrationWriter):
    """
    Write nicely formatted migration files.

    Uses `Black <https://black.readthedocs.io/en/stable/>`_ and `isort
    <https://pycqa.github.io/isort/>`_ with settings compatible with :ref:`pytest-logikal
    <pytest-logikal:index:Getting Started>`.

    .. note:: Requires :ref:`pytest-logikal <pytest-logikal:index:Installation>` to be installed.
    """
    def as_string(self) -> str:
        """
        Return a string of the nicely formatted file contents.
        """
        import black
        import isort
        from pytest_logikal.black import get_mode as black_mode
        from pytest_logikal.isort import get_config as isort_config
        from pytest_logikal.utils import get_ini_option

        max_line_length = get_ini_option('max_line_length')
        code = super().as_string()
        code = black.format_str(code, mode=black_mode(max_line_length))
        return isort.api.sort_code_string(code, **isort_config(max_line_length=max_line_length))



# pylint: disable=no-value-for-parameter
from calendar import month_name
from datetime import datetime
from logging import getLogger
from typing import Any, Dict, Mapping, Sequence

from django.utils.html import escape
from django.utils.safestring import SafeString, mark_safe
from pybtex.database import BibliographyData, parse_file
from pybtex.richtext import Symbol, Text
from pybtex.style.formatting import BaseStyle
from pybtex.style.template import (
    field, first_of, href, join, names, optional, optional_field, tag, together,
)

logger = getLogger(__name__)


def format_url(text: str) -> Text:
    return Text(text.replace(' ', ''))


def format_month(text: str) -> Text:
    return Text(month_name[int(text)])


def format_date(text: str) -> Text:
    text_date = datetime.strptime(text, '%Y-%m-%d').strftime('%B %-d, %Y')
    return Symbol('nbsp').join(Text(text_date).split(' '))


def dash(text: Text) -> Text:
    return Text(Symbol('ndash')).join(text.split('-'))


url = href()[  # note: add external=True when the next version of pybtex is released
    field('url', apply_func=format_url, raw=True),  # href
    field('url', apply_func=format_url, raw=True),  # text content
]

date = first_of[
    optional_field('date', apply_func=format_date, raw=True),
    optional[together(last_tie=True)[
        field('month', apply_func=format_month, raw=True), field('year')
    ]],
    optional_field('year'),
]


class WebStyle(BaseStyle):  # type: ignore[misc]
    @staticmethod
    def separated_names(field_name: str) -> Text:
        return names(field_name, sep=', ', sep2=' and ', last_sep=' and ')

    def get_article_template(self, *args: Any, **kwargs: Any) -> Text:
        return join(sep=', ')[
            self.separated_names('author'),
            tag('em')[field('title')],
            field('journaltitle'),
            optional[
                together(last_tie=True)['Volume', field('volume')],
                optional['(', field('number'), ')'],
            ],
            optional[together(last_tie=True)['Issue', field('issue')]],
            optional_field('note'),
            date,
            optional[together(last_tie=True)['Pages', field('pages', apply_func=dash)]],
            optional[url],
        ]

    def get_book_template(self, *args: Any, **kwargs: Any) -> Text:
        return join(sep=', ')[
            self.separated_names('author'),
            tag('em')[field('title')],
            optional_field('publisher'),
            optional_field('address'),
            date,
            optional[together(last_tie=True)['ISBN', field('isbn')]],
            optional[url],
        ]

    def get_online_template(self, *args: Any, **kwargs: Any) -> Text:
        return join(sep=', ')[
            optional[self.separated_names('author')],
            optional_field('title'),
            date,
            optional_field('note'),
            url,
        ]

    def get_inproceedings_template(self, *args: Any, **kwargs: Any) -> Text:
        return join(sep=', ')[
            self.separated_names('author'),
            tag('em')[field('title')],
            field('booktitle'),
            optional_field('series'),
            optional_field('publisher'),
            optional_field('address'),
            date,
            optional[together(last_tie=True)['ISBN', field('isbn')]],
            optional[together(last_tie=True)['Pages', field('pages', apply_func=dash)]],
            optional[url],
        ]

    def get_misc_template(self, *args: Any, **kwargs: Any) -> Text:
        return join(sep=', ')[
            optional[self.separated_names('author')],
            optional[tag('em')[field('title')]],
            optional[tag('em')[field('subtitle')]],
            date,
            optional_field('note'),
            optional[url],
        ]


class Bibliography:
    data: Dict[str, BibliographyData] = {}

    def __init__(self, name: str):
        """
        Render bibliography citations and references.

        Args:
            name: The name of the bibliography configuration to use.

        """
        self._name = name
        self._data = Bibliography.data[name]
        self._index = 0
        self._references: Dict[str, BibliographyData] = {}

    @classmethod
    def add_bibliographies(cls, bibliographies: Mapping[str, str]) -> None:
        """
        Parse and store the provided bibliographies.

        Args:
            bibliographies: A name to path mapping of the bibliographies to read.

        """
        for name, path in bibliographies.items():
            if name not in cls.data:
                logger.debug(f'Loading bibliography "{path}"')
                cls.data[name] = parse_file(path, bib_format='bibtex').lower()

    def cite(self, name: str) -> SafeString:
        """
        Render a citation to the given entry and store it in the references.

        Args:
            name: The name of the entry.

        """
        if (name := escape(name).lower()) in self._references:  # pybtex is case insensitive
            raise RuntimeError(f'Reference "{name}" has already been used')
        try:
            self._references[name] = self._data.entries[name]
        except KeyError as error:
            raise RuntimeError(f'Reference "{name}" not found in "{self._name}"') from error

        self._index += 1
        return mark_safe(  # nosec: name is escaped and index is an internal integer
            f'<a href="#ref-{name}" id="cite-{name}" class="cite">'
            f'[<span>{self._index}</span>]</a>'
        )

    # Note: this function is only necessary because it is not possible to configure pybtex links to
    # open in a new tab (feature is implemented but not released)
    @staticmethod
    def _render_reference(style: WebStyle, name: str, entry: BibliographyData) -> str:
        reference = str(style.format_entry(name, entry).text.render_as('html'))
        return reference.replace('<a ', '<a target="_blank" ')

    def references(self, classes: Sequence[str] = ('references', )) -> SafeString:
        """
        Render the stored references as an ordered, numbered list.

        Args:
            classes: The classes to add to the ordered list element.

        """
        style = WebStyle(name_style='plain')
        items = '\n'.join(
            f'<li id="ref-{name}"><a href="#cite-{name}" class="cite up">&uarr;</a>'
            f'{self._render_reference(style=style, name=name, entry=entry)}</li>'
            for name, entry in self._references.items()
        )
        return mark_safe(  # nosec: name and entries are escaped
            f'<ol class="{" ".join(classes)}">\n{items}\n</ol>'
        )



from enum import Enum

from django.db.models import Case, TextChoices, Value, When
from django.db.models.expressions import Expression


class OrderedTextChoices(TextChoices):
    @classmethod
    def order(cls, field_name: str) -> Expression:
        return Case(*(
            When(**{field_name: attribute.label}, then=Value(index))  # type: ignore
            for index, (name, attribute) in enumerate(cls.__dict__.items())
            if isinstance(attribute, Enum)
        ))



from django_logikal.settings import Settings
from django_logikal.settings.common.dev import CommonDevSettings
from django_logikal.settings.static_site.base import BaseSettings


class DevSettings(CommonDevSettings, BaseSettings):
    """
    Standard settings for developing static sites.
    """
    @staticmethod
    def apply(settings: Settings) -> None:
        pass



from os import environ, getenv

PREFIX = 'DJANGO_LOGIKAL_'


def get_option(option: str, default: str = '') -> str:
    return getenv(f'{PREFIX}{option.upper()}', default=default)


def set_option(option: str, value: str = '1') -> None:
    environ[f'{PREFIX}{option.upper()}'] = value


def option_is_set(option: str) -> bool:
    return get_option(option) == '1'


def is_dev() -> bool:
    return option_is_set('dev_run') or option_is_set('testing')



from django_logikal.env import option_is_set
from django_logikal.settings import Settings
from django_logikal.settings.common.dev import CommonDevSettings
from django_logikal.settings.dynamic_site.base import BaseSettings


class DevSettings(CommonDevSettings, BaseSettings):
    """
    Standard settings for developing dynamic sites.
    """
    @staticmethod
    def apply(settings: Settings) -> None:
        settings['INSTALLED_APPS'].insert(
            settings['INSTALLED_APPS'].index('django.contrib.staticfiles'),
            'whitenoise.runserver_nostatic',
        )

        if not option_is_set('send_emails'):
            settings['EMAIL_BACKEND'] = 'django.core.mail.backends.console.EmailBackend'



from django_logikal.env import set_option
from django_logikal.logging import logging_config
from django_logikal.settings import Settings, SettingsUpdate


class CommonTestingSettings(SettingsUpdate):
    LOGGING = logging_config(console=False)  # logs are captured by pytest already

    @staticmethod
    def apply(settings: Settings) -> None:
        settings['STORAGES']['staticfiles'] = {
            'BACKEND': 'django.contrib.staticfiles.storage.StaticFilesStorage',
        }

        # Custom settings
        set_option('testing')



from itertools import chain
from logging import getLogger
from typing import Any

from django.http import HttpRequest, HttpResponse
from django.shortcuts import render
from django.utils.safestring import mark_safe
from pygments import highlight
from pygments.formatters.html import HtmlFormatter
from pygments.lexers.html import HtmlLexer

from django_logikal.middleware import Middleware

logger = getLogger(__name__)


class ValidationMiddleware(Middleware):
    """
    Validate HTML responses for all requests.

    .. note:: Requires :ref:`pytest-logikal <pytest-logikal:index:Installation>` to be installed.
    """
    def __init__(self, *args: Any, **kwargs: Any):
        from pytest_logikal.validator import Validator  # pylint: disable=import-outside-toplevel

        super().__init__(*args, **kwargs)
        self._validator = Validator()

    def __call__(self, request: HttpRequest) -> HttpResponse:
        response = super().__call__(request)
        resolver_match = getattr(request, 'resolver_match', None)
        app_name = getattr(resolver_match, 'app_name', None)
        if (
            not response.headers['Content-Type'].startswith('text/html')
            or (response.status_code != 200 and app_name != 'error')
            or app_name in {'admin', 'djdt'}  # admin and Django debug toolbar responses
        ):
            return response

        logger.info(f'Validating HTML response for "{request.path}"')
        content = response.content.decode()
        if errors := self._validator.errors(content=content):
            template = 'django_logikal/validation_error.html.j'
            lines = list(chain.from_iterable(
                range(error.first_line, error.last_line + 1)
                for error in errors if error.first_line and error.last_line
            ))
            context = {
                # These are safe strings
                'code_styles': mark_safe(HtmlFormatter(style='manni').get_style_defs()),  # nosec
                'source': mark_safe(highlight(  # nosec
                    code=content,
                    formatter=HtmlFormatter(linenos=True, hl_lines=lines, wrapcode=True),
                    lexer=HtmlLexer(),
                )),
                'errors': errors,
            }
            response = render(request, template_name=template, context=context, status=500)
            response.reason_phrase = 'Invalid HTML'
        return response



from typing import Callable

from django.http import HttpRequest, HttpResponse


class Middleware:
    def __init__(self, get_response: Callable[[HttpRequest], HttpResponse]):
        self.get_response = get_response

    def __call__(self, request: HttpRequest) -> HttpResponse:
        return self.get_response(request)



import logging
import warnings
from time import localtime, strftime
from typing import Any, Callable, Mapping, Optional, Type

from django.conf import settings
from google.cloud import logging as cloud_logging
from termcolor import colored


class ConsoleLogFormatter(logging.Formatter):
    _STYLES: Mapping[str, Any] = {
        'timestamp': {'color': 'magenta'},
        'level': {
            'DEBUG': {'color': 'blue'},
            'INFO': {'color': 'white', 'attrs': ['dark']},
            'WARNING': {'color': 'red'},
            'ERROR': {'color': 'red', 'attrs': ['bold']},
            'CRITICAL': {'color': 'red', 'attrs': ['bold']},
        },
        'place': {'color': 'white', 'attrs': ['dark']},
        'status_code': {
            'HTTP_INFO': {'attrs': ['bold']},
            'HTTP_NOT_MODIFIED': {'color': 'cyan'},
            'HTTP_REDIRECT': {'color': 'green'},
            'HTTP_NOT_FOUND': {'color': 'yellow'},
            'HTTP_BAD_REQUEST': {'color': 'red', 'attrs': ['bold']},
            'HTTP_SERVER_ERROR': {'color': 'magenta', 'attrs': ['bold']},
        }
    }

    def colored_message(self, message: str, status_code: Optional[int]) -> str:
        if not status_code or (200 <= status_code < 300):
            return message
        if 100 <= status_code < 200:
            style = 'HTTP_INFO'
        elif status_code == 304:
            style = 'HTTP_NOT_MODIFIED'
        elif 300 <= status_code < 400:
            style = 'HTTP_REDIRECT'
        elif status_code == 404:
            style = 'HTTP_NOT_FOUND'
        elif 400 <= status_code < 500:
            style = 'HTTP_BAD_REQUEST'
        else:
            style = 'HTTP_SERVER_ERROR'
        return colored(message, **self._STYLES['status_code'][style])

    def format(self, record: logging.LogRecord) -> str:
        timestamp = strftime('%Y-%m-%d %H:%M:%S', localtime(record.created))
        timestamp = colored(f'{timestamp}.{record.msecs:03.0f}', **self._STYLES['timestamp'])
        level = record.levelname
        level = colored(level, **(self._STYLES['level'].get(level, self._STYLES['level']['INFO'])))
        place = f'({record.name}:{record.lineno})'
        place = colored(place, **self._STYLES['place'])
        message = record.getMessage().strip()
        message = self.colored_message(message, status_code=getattr(record, 'status_code', None))

        if exception := (getattr(record, 'exc_info', '') or ''):
            exception = '\n' + self.formatException(record.exc_info)  # type: ignore[arg-type]
        if stack := (getattr(record, 'stack_info', '') or ''):
            stack = '\n' + self.formatStack(record.stack_info)  # type: ignore[arg-type]

        return f'{timestamp} {level} {message} {place}{exception}{stack}'

    @staticmethod
    def warning_formatter() -> Callable[..., str]:
        def formatwarning(
            message: Warning, category: Type[Warning], filename: str, lineno: int,
            *_args: Any, **_kwargs: Any,
        ) -> str:
            cleaned_message = str(message).replace('\n', ' ').replace('  ', ' ').strip()
            return f'{category.__name__}: {cleaned_message} ({filename}:{lineno})'
        return formatwarning


def logging_config(
    log_level: str = 'INFO',
    console: bool = True,
    cloud: bool = False,
) -> Mapping[str, Any]:
    config: Mapping[str, Any] = {
        'version': 1,
        'disable_existing_loggers': False,
        'formatters': {'console': {'()': ConsoleLogFormatter}},
        'handlers': {
            'console': {'class': 'logging.StreamHandler', 'formatter': 'console', 'filters': []},
        },
        'loggers': {
            'django': {'propagate': True, 'handlers': []},
            'django.db': {'propagate': True, 'handlers': [], 'level': log_level},
        },
        'root': {'level': log_level, 'handlers': []},
    }

    if console:
        config['root']['handlers'].append('console')
        warnings.formatwarning = ConsoleLogFormatter.warning_formatter()

    if cloud:
        from stormware.google.auth import GCPAuth  # pylint: disable=import-outside-toplevel

        auth = GCPAuth()
        client = cloud_logging.Client(  # type: ignore[no-untyped-call]
            project=auth.project_id(),
            credentials=auth.credentials(),
        )
        config['handlers']['cloud'] = {
            'class': 'google.cloud.logging.handlers.CloudLoggingHandler',
            'client': client,
            'name': f'django-{settings.SETTINGS_MODULE.rsplit(".", 1)[-1].replace("_", "-")}',
        }
        config['root']['handlers'].append('cloud')

    return config



import re

from django.utils.html import escape
from django.utils.safestring import SafeString, mark_safe
from django.utils.text import slugify as django_slugify

SLUGIFY_TRANSLATION = str.maketrans({
    '&': 'and',
    'ø': 'o',
    'æ': 'ae',
})


def join_lines(text: str, spacer: bool = False) -> str:
    """
    Replace new lines with spaces.

    Args:
        text: The text to use.
        spacer: Whether to add a space to the beginning when the text is not empty.

    """
    joined = re.sub(' +', ' ', text.replace('\n', ' ')).strip()
    prefix = ' ' if joined and spacer else ''
    return f'{prefix}{joined}'


def slugify(text: str, use_underscore: bool = False) -> str:
    """
    Convert a string to a URL slug.
    """
    slugified = django_slugify(text.translate(SLUGIFY_TRANSLATION))
    return slugified if not use_underscore else slugified.replace('-', '_')


def unslugify(text: str) -> str:
    """
    Convert a URL slug to a string.
    """
    return text.replace('and', '&').replace('-', ' ')


def wrap(text: str) -> SafeString:
    """
    Replace spaces with line breaks.
    """
    text = escape(text).replace(' &amp; ', '&nbsp;&amp;&nbsp;').replace(' ', '<br>')
    return mark_safe(text)  # nosec: text is escaped


def nowrap(text: str) -> SafeString:
    """
    Replace spaces with non-breaking spaces.
    """
    return mark_safe(escape(text).replace(' ', '&nbsp;'))  # nosec: text is escaped



from typing import Any, List

from django.apps import AppConfig
from django.conf import global_settings, settings
from django.core.checks import Error, register
from django.db.migrations import writer
from django.template.loader import get_template

from django_logikal.migrations import FormattedMigrationWriter


class DjangoLogikalConfig(AppConfig):
    name = 'django_logikal'
    verbose_name = 'django-logikal'

    def ready(self) -> None:
        # Load bibliographies
        if bibliographies := getattr(settings, 'BIBLIOGRAPHIES', None):
            from django_logikal.bibliography import (  # pylint: disable=import-outside-toplevel
                Bibliography,
            )

            Bibliography.add_bibliographies({
                # Note: both Django and Jinja2 templates have an origin attribute, so this is fine
                name: get_template(path).origin.name  # type: ignore[attr-defined]
                for name, path in bibliographies.items()
            })

        # Format migrations
        writer.MigrationWriter = FormattedMigrationWriter  # type: ignore[misc]


@register
def check_settings(*_args: Any, **_kwargs: Any) -> List[Error]:
    errors = []

    # Check ROOT_URLCONF
    if not getattr(settings, 'ROOT_URLCONF', None):
        errors.append(Error(
            msg='The ROOT_URLCONF setting must be specified', obj=str(settings),
            id='django_logikal.E001',
        ))

    # Check AUTH_USER_MODEL
    if (
        'django.contrib.auth' in getattr(settings, 'INSTALLED_APPS', [])
        and getattr(settings, 'AUTH_USER_MODEL', None) == global_settings.AUTH_USER_MODEL
    ):
        errors.append(Error(
            msg='A custom AUTH_USER_MODEL setting must be provided', obj=str(settings),
            id='django_logikal.E002',
        ))

    # Check EMAIL_SUBJECT_PREFIX
    if (
        'anymail' in getattr(settings, 'INSTALLED_APPS', [])
        and getattr(settings, 'EMAIL_SUBJECT_PREFIX') == global_settings.EMAIL_SUBJECT_PREFIX
    ):
        errors.append(Error(
            msg='A custom EMAIL_SUBJECT_PREFIX setting must be provided', obj=str(settings),
            id='django_logikal.E003'
        ))

    return errors



try:
    import zoneinfo  # type: ignore[import-not-found]
except ImportError:
    from backports import zoneinfo  # note: this is built-in after Python 3.9+
from django.utils import timezone, translation
from jinja2 import nodes
from jinja2.ext import Extension
from jinja2.parser import Parser
from jinja2.runtime import Macro


class LanguageExtension(Extension):
    tags = {'language'}

    def parse(self, parser: Parser) -> nodes.Node:
        lineno = next(parser.stream).lineno
        args = [parser.parse_expression()]
        body = parser.parse_statements(tuple(['name:endlanguage']), drop_needle=True)
        block = nodes.CallBlock(self.call_method('_override_language', args), [], [], body)
        return block.set_lineno(lineno)

    @staticmethod
    def _override_language(language_code: str, caller: Macro) -> str:
        with translation.override(language_code):
            return caller()


class TimeZoneExtension(Extension):
    tags = {'timezone'}

    def parse(self, parser: Parser) -> nodes.Node:
        lineno = next(parser.stream).lineno
        args = [parser.parse_expression()]
        body = parser.parse_statements(tuple(['name:endtimezone']), drop_needle=True)
        block = nodes.CallBlock(self.call_method('_override_timezone', args), [], [], body)
        return block.set_lineno(lineno)

    @staticmethod
    def _override_timezone(zone_id: str, caller: Macro) -> str:
        current_timezone = timezone.get_current_timezone()
        try:
            timezone.activate(zoneinfo.ZoneInfo(zone_id))
            return caller()
        finally:
            timezone.activate(current_timezone)



from typing import Any, Dict

from stormware.amazon.auth import AWSAuth

from django_logikal.settings import Settings
from django_logikal.settings.common.base import CommonBaseSettings


class BaseSettings(CommonBaseSettings):
    MIDDLEWARE = [
        'django.middleware.security.SecurityMiddleware',  # required for HTTPS redirection & others
        'whitenoise.middleware.WhiteNoiseMiddleware',  # static file serving
        'django.contrib.sessions.middleware.SessionMiddleware',  # required by Django admin
        'django.middleware.locale.LocaleMiddleware',  # required for localization
        'django.middleware.common.CommonMiddleware',  # performs URL rewriting and sets headers
        'django.middleware.csrf.CsrfViewMiddleware',  # adds hidden form fields to POST forms
        'django.contrib.auth.middleware.AuthenticationMiddleware',  # required by Django admin
        'django_logikal.security.LoginRequiredByDefaultMiddleware',  # force login by default
        'django.contrib.messages.middleware.MessageMiddleware',  # required by Django admin
        'django.contrib.sites.middleware.CurrentSiteMiddleware',  # adds site attribute to requests
        'django.middleware.clickjacking.XFrameOptionsMiddleware',  # defense against clickjacking
        'csp.middleware.CSPMiddleware',  # adds the Content-Security-Policy header
    ]

    # Security
    SECURE_REFERRER_POLICY = ['same-origin', 'origin-when-cross-origin']
    SESSION_COOKIE_AGE = 7 * 24 * 60 * 60  # 7 days (default: 14 days)
    CSP_DEFAULT_SRC = "'self'"  # the single quotes are important
    CSP_BLOCK_ALL_MIXED_CONTENT = True
    CSP_INCLUDE_NONCE_IN = ['default-src']

    # Internationalization
    LANGUAGE_COOKIE_NAME = 'language'
    LANGUAGE_COOKIE_SAMESITE = 'Lax'

    # Authentication
    LOGIN_URL = 'admin:login'
    LOGIN_REDIRECT_URL = 'admin:index'
    LOGOUT_REDIRECT_URL = 'admin:login'
    AUTH_MIN_PASSWORD_LENGTH = 8
    AUTH_PASSWORD_VALIDATORS = [{
        'NAME': 'django.contrib.auth.password_validation.MinimumLengthValidator',
        'OPTIONS': {'min_length': AUTH_MIN_PASSWORD_LENGTH},
    }]
    PASSWORD_RESET_TIMEOUT = 1 * 24 * 60 * 60  # 1 day (default: 3 days)

    # Email
    EMAIL_BACKEND = 'anymail.backends.amazon_ses.EmailBackend'
    EMAIL_TIMEOUT = 10  # default: infinite
    ANYMAIL: Dict[str, Any] = {
        'REQUESTS_TIMEOUT': 10,  # default: 30s
        'AMAZON_SES_SESSION_PARAMS': {'profile_name': AWSAuth().profile()},
    }

    # Static files
    STORAGES = {**CommonBaseSettings.STORAGES, **{
        'staticfiles': {'BACKEND': 'whitenoise.storage.CompressedManifestStaticFilesStorage'},
    }}
    WHITENOISE_KEEP_ONLY_HASHED_FILES = True

    @staticmethod
    def apply(settings: Settings) -> None:
        settings['INSTALLED_APPS'] += [
            'django.contrib.admin',
            'django.contrib.messages',  # required by Django admin
            'django.contrib.sessions',  # required by Django admin
            'anymail',
        ]



import re
from typing import Any, Dict

from django.conf import settings
from django.template import Origin, TemplateDoesNotExist
from django.template.backends.jinja2 import Jinja2
from django.utils import translation
from jinja2.environment import Environment
from jinja2.runtime import StrictUndefined

from django_logikal.templates import filters, functions, tests


class JinjaTemplates(Jinja2):
    app_dirname = 'templates'

    def __init__(self, params: Dict[str, Any]):
        params = params.copy()
        self._extension = params['OPTIONS'].pop('match_extension', '.j')
        environment_path = f'{environment.__module__}.{environment.__qualname__}'
        params['OPTIONS'].setdefault('environment', environment_path)
        params['OPTIONS'].setdefault('undefined', StrictUndefined)
        params['OPTIONS'].setdefault('trim_blocks', True)
        params['OPTIONS'].setdefault('lstrip_blocks', True)
        # See https://github.com/pallets/jinja/issues/178
        # See https://github.com/pallets/jinja/pull/1456
        # params['OPTIONS'].setdefault('indent_blocks', True)
        params['OPTIONS'].setdefault('extensions', [
            'jinja2.ext.i18n',
            'csp.extensions.NoncedScript',
            'django_logikal.templates.extensions.LanguageExtension',
            'django_logikal.templates.extensions.TimeZoneExtension',
        ])
        params['OPTIONS'].setdefault('context_processors', [
            'django_logikal.templates.processors.add_messages',
        ])
        super().__init__(params)

    def get_template(self, template_name: str) -> Any:
        if template_name.endswith(self._extension):
            return super().get_template(template_name)
        origin = Origin(name=template_name, loader=self.env.loader)
        error = f'Skipping template search as the template extension is not "{self._extension}"'
        raise TemplateDoesNotExist(template_name, tried=[(origin, error)], backend=self)


def environment(**options: Dict[str, Any]) -> Environment:
    options = {option: value for option, value in options.items() if option != 'autoescape'}
    env = Environment(**options, autoescape=True)  # type: ignore[arg-type]
    env.policies.update({'ext.i18n.trimmed': True})
    # Note: we might need to use a different gettext
    # (see https://code.djangoproject.com/ticket/34602)
    env.install_gettext_callables(  # type: ignore[attr-defined] # pylint: disable=no-member
        gettext=translation.gettext,
        ngettext=translation.ngettext,
        newstyle=True,
        pgettext=translation.pgettext,
        npgettext=translation.npgettext,
    )
    env.filters.update({
        # Built-in functions
        'dir': dir,
        'getattr': getattr,
        'repr': repr,
        'type': type,
        # Built-in types
        'str': str,
        # Other utilities
        'join_lines': filters.join_lines,
        'slugify': filters.slugify,
        'unslugify': filters.unslugify,
        'wrap': filters.wrap,
        'nowrap': filters.nowrap,
    })
    env.tests.update({
        'startswith': tests.startswith,
    })
    env.globals.update({
        # Django objects
        'settings': settings,
        'filters': env.filters,
        'tests': env.tests,
        # Libraries
        're': re,
        # Django utilities
        'context': functions.context,
        'static': functions.static,
        'static_path': functions.static_path,
        'include_static': functions.include_static,
        'url': functions.url,
        'url_name': functions.url_name,
        'language': functions.language,
        'format': functions.format,
        # Other utilities
        'cwd': functions.cwd,
        'now': functions.now,
        'bibliography': functions.bibliography,
    })
    return env



from typing import Dict, List, Union

from django.contrib.messages import get_messages
from django.contrib.messages.storage.base import BaseStorage, Message
from django.http.request import HttpRequest


def add_messages(request: HttpRequest) -> Dict[str, Union[List[Message], BaseStorage]]:
    return {'messages': get_messages(request=request)}



"""
Execute a Django management command.
"""
import sys
from os import environ
from typing import Sequence

from django.core.management import execute_from_command_line
from logikal_utils.project import tool_config


def main(args: Sequence[str] = tuple(sys.argv[1:])) -> int:
    sys.path.insert(0, '.')
    environ.setdefault(
        'DJANGO_SETTINGS_MODULE',
        tool_config('django_logikal')['DJANGO_SETTINGS_MODULE'],
    )
    execute_from_command_line(['manage'] + list(args))
    return 0



from functools import wraps
from typing import Any, Callable

from django.http import HttpResponseBase
from django.http.response import HttpResponseNotFound, HttpResponseServerError
from django.views import View, defaults, generic

ViewFunction = Callable[..., HttpResponseBase]


class PublicViewMixin:
    """
    Mark a class-based view public.
    """


class PublicView(PublicViewMixin, View):
    # Note: we re-define init to override the inherited class docstring
    def __init__(self, *args: Any, **kwargs: Any):  # pylint: disable=useless-parent-delegation
        """
        Represent a public view.
        """
        super().__init__(*args, **kwargs)  # pragma: no cover


def public(view: ViewFunction) -> ViewFunction:
    """
    Mark a view function public.
    """
    @wraps(view)
    def view_wrapper(*args: Any, **kwargs: Any) -> HttpResponseBase:
        return view(*args, **kwargs)
    view_wrapper.public_view = True  # type: ignore[attr-defined]
    return view_wrapper


def redirect_to(viewname: str) -> Callable[..., HttpResponseBase]:
    """
    Redirect to the given view.
    """
    return generic.RedirectView.as_view(pattern_name=viewname, query_string=True)


@public
def page_not_found_view(*args: Any, **kwargs: Any) -> HttpResponseNotFound:
    """
    Public view rendered from the ``404.html.j`` template.
    """
    kwargs.setdefault('template_name', '404.html.j')
    kwargs.setdefault('exception', None)
    return defaults.page_not_found(*args, **kwargs)


@public
def server_error_view(*args: Any, **kwargs: Any) -> HttpResponseServerError:
    """
    Public view rendered from the ``500.html.j`` template.
    """
    kwargs.setdefault('template_name', '500.html.j')
    return defaults.server_error(*args, **kwargs)


#: Standard error handler views.
ERROR_HANDLERS = {
    400: server_error_view,
    403: page_not_found_view,
    404: page_not_found_view,
    500: server_error_view,
}



"""
Manage translation message catalogs.
"""
import logging
import os
import sys
from importlib import metadata
from pathlib import Path
from typing import Any, Optional, Sequence, cast

import django
from babel.messages.frontend import CommandLineInterface
from django.core.management.base import BaseCommand, CommandError, CommandParser
from logikal_utils.project import PYPROJECT, tool_config

TEMPLATE_HEADER = """
# Translation template for project "PROJECT"
# Copyright 2023 ORGANIZATION
"""
DEFAULT_WIDTH = 76
DOMAINS = ('django', 'djangojs')


class App:
    def __init__(self, name: str):
        self.name = name
        app_config = django.apps.apps.get_app_config(name)
        self.path = Path(os.path.relpath(app_config.path, os.getcwd()))


class Command(BaseCommand):
    help = __doc__

    def add_arguments(self, parser: CommandParser) -> None:
        group = parser.add_argument_group(title='actions').add_mutually_exclusive_group()
        group.add_argument('-i', '--init', action='store_true',
                           help='Initialize new message catalogs.')
        group.add_argument('-u', '--update', action='store_true',
                           help='Update message catalogs.')
        group.add_argument('-c', '--compile', action='store_true',
                           help='Compile message catalogs.')
        parser.add_argument('app', nargs='*', help='The app to translate.')
        parser.add_argument('-l', '--locale', help='Locale for the new catalogs.')
        parser.add_argument('-o', '--output', type=Path, help='Output path to use.')

    def handle(self, *_args: Any, **options: Any) -> None:
        def remove_config_info(record: logging.LogRecord) -> bool:
            if record.msg.lower() == 'extracting messages from %s%s':
                record.args = (record.args[0], '')  # type: ignore[index]
            return True

        logging.getLogger().handlers = []  # to avoid double logging
        logging.getLogger('babel').addFilter(remove_config_info)

        default_apps = tool_config('django_logikal').get('translate', {}).get('apps', [])
        if not (apps := [App(app) for app in options.get('app') or default_apps]):
            raise CommandError('At least one app name must be provided')

        output = cast(Optional[Path], options.get('output'))  # wrong typing

        if options.get('init'):
            if not (locale := cast(Optional[str], options.get('locale'))):
                raise CommandError('The locale must be provided')
            return self.action_init(apps=apps, locale=locale, output=output)
        if options.get('update'):
            return self.action_update(apps=apps, output=output)
        if options.get('compile'):
            return self.action_compile(apps=apps, output=output)

        raise CommandError('An action must be provided')

    def _extract(self, app: App, output: Optional[Path]) -> None:
        self.stdout.write()
        locale_path = (output or app.path) / 'locale'
        locale_path.mkdir(parents=True, exist_ok=True)
        for domain in DOMAINS:
            CommandLineInterface().run([  # type: ignore[no-untyped-call]
                sys.argv[0], 'extract',
                # Configuration
                '--mapping-file', Path(__file__).parents[2] / f'babel/{domain}.ini',
                '--width', DEFAULT_WIDTH,
                '--add-comments', 'Translators:',
                # Metadata
                '--project', app.name.replace('_', '-'),
                '--version', f'v{metadata.version(PYPROJECT["project"]["name"])}',
                '--msgid-bugs-address', tool_config('django_logikal')['translate']['contact'],
                '--copyright-holder', PYPROJECT['project']['authors'][0]['name'],
                '--header-comment', TEMPLATE_HEADER,
                # Input and output file
                '--output-file', str(locale_path / f'{domain}.pot'),
                str(app.path),
            ])
        self.stdout.write()

    def action_init(self, apps: Sequence[App], locale: str, output: Optional[Path]) -> None:
        app_names = ', '.join(self.style.ERROR(app.name) for app in apps)
        self.stdout.write(f'Creating message catalogs for {app_names}\n')

        # Checking for overwrites
        output_files = {
            app.name: {
                domain: (output or app.path) / 'locale' / locale / f'LC_MESSAGES/{domain}.po'
                for domain in DOMAINS
            }
            for app in apps
        }
        existing_files = {
            app_name: [file for file in domain_files.values() if file.exists()]
            for app_name, domain_files in output_files.items()
        }
        if any(existing_files.values()):
            warning = '\nThis will OVERWRITE the following message catalog files:'
            self.stdout.write(self.style.WARNING(warning))
            for app_name, files in existing_files.items():
                indent = 0
                if len(existing_files) > 1:
                    indent += 2
                    self.stdout.write(indent * ' ' + self.style.ERROR(app_name))
                self.stdout.write('\n'.join((indent + 2) * ' ' + str(file) for file in files))
            self.stdout.write('\nAre you sure you want to do this?')
            prompt = 'Type \'yes\' to continue, or \'no\' to cancel: '
            if input(prompt) != 'yes':
                raise CommandError('Cancelled')

        # Creating message catalog
        for app in apps:
            if len(apps) > 1:
                self.stdout.write()
                self.stdout.write(f'Creating message catalogs for {self.style.ERROR(app.name)}\n')
            self._extract(app=app, output=output)
            for domain in DOMAINS:
                CommandLineInterface().run([  # type: ignore[no-untyped-call]
                    sys.argv[0], 'init',
                    '--width', DEFAULT_WIDTH,
                    '--locale', locale,
                    '--domain', domain,
                    '--input-file', str((output or app.path) / f'locale/{domain}.pot'),
                    '--output-file', str(output_files[app.name][domain]),
                ])

        self.stdout.write(self.style.SUCCESS('\nMessage catalog successfully created'))

    def action_update(self, apps: Sequence[App], output: Optional[Path]) -> None:
        for app in apps:
            self.stdout.write(f'Updating catalogs for {self.style.ERROR(app.name)}\n')
            self._extract(app=app, output=output)
            for domain in DOMAINS:
                CommandLineInterface().run([  # type: ignore[no-untyped-call]
                    sys.argv[0], 'update',
                    '--width', DEFAULT_WIDTH,
                    '--domain', domain,
                    '--input-file', str((output or app.path) / f'locale/{domain}.pot'),
                    '--output-dir', str((output or app.path) / 'locale'),
                ])
            self.stdout.write()
        self.stdout.write(self.style.SUCCESS('Message catalogs successfully updated'))

    def action_compile(self, apps: Sequence[App], output: Optional[Path]) -> None:
        for app in apps:
            self.stdout.write(f'Compiling message catalogs for {self.style.ERROR(app.name)}\n\n')
            CommandLineInterface().run([  # type: ignore[no-untyped-call]
                sys.argv[0], 'compile',
                '--domain', ' '.join(DOMAINS),
                '--directory', str((output or app.path) / 'locale'),
                '--statistics',
            ])
            self.stdout.write()
        self.stdout.write(self.style.SUCCESS('Message catalogs successfully compiled'))



from django_logikal.settings.common.production import CommonProductionSettings
from django_logikal.settings.dynamic_site.base import BaseSettings


class ProductionSettings(BaseSettings, CommonProductionSettings):
    """
    Standard production settings for dynamic sites.

    .. note:: Secrets will be loaded from Google Secret Manager during import time. In particular,
        the secret key is loaded from the ``SECRET_KEY_PATH`` setting key (defaults to
        ``{project}-website-secret-key``, where ``{project}`` is the project name in the
        ``pyproject.toml`` file), and the database configuration is loaded from the
        ``DATABASE_SECRETS_PATH`` setting key (defaults to ``{project}-website-database-secrets``),
        which must contain a JSON string with keys ``hostname``, ``port``, ``database``,
        ``username`` and ``password``.
    """



from typing import Any, List, Mapping, Optional, Sequence, Tuple, Type, Union

import debug_toolbar
from django.contrib import admin
from django.contrib.sitemaps import Sitemap
from django.contrib.sitemaps.views import sitemap
from django.urls import URLPattern, URLResolver, include, path

from django_logikal.env import is_dev, option_is_set
from django_logikal.views import ERROR_HANDLERS, public

URLType = Union[URLResolver, URLPattern]
IncludeType = Tuple[Sequence[URLType], Optional[str], Optional[str]]


def admin_urls() -> IncludeType:
    """
    Return URLs for the :mod:`django.contrib.auth` app.
    """
    urls = admin.site.urls
    for url in urls[0]:
        if getattr(url, 'name', None) == 'login' and url.callback:
            url.callback = public(url.callback)
            break
    return urls


def error_urls() -> IncludeType:
    """
    Return URLs for the standard error handler pages.
    """
    return include(([
        path(f'{code}/', view, name=str(code))
        for code, view in ERROR_HANDLERS.items()
    ], 'error'))


def debug_toolbar_urls() -> IncludeType:  # pragma: no cover, tested in subprocess
    """
    Return URLs for the Django debug toolbar.
    """
    urls = list(debug_toolbar.toolbar.DebugToolbar.get_urls())
    for url in urls:
        for url_item in getattr(url, 'url_patterns', [url]):
            url_item.callback = public(url_item.callback)
    return include((urls, debug_toolbar.APP_NAME))


def utility_paths(
    sitemaps: Optional[Mapping[str, Union[Sitemap[Any], Type[Sitemap[Any]]]]] = None,
    static: bool = False,
) -> List[URLType]:
    """
    Return the common utility paths.

    Includes the admin paths, sitemap and robots paths and the Django debug toolbar paths (when
    appropriate). Also includes the standard error paths when running locally.
    """
    # Note: we have to import late for documentation building to succeed
    from robots.views import rules_list  # pylint: disable=import-outside-toplevel

    paths: List[URLType] = []
    universal_path = path  # included for both static and dynamic sites
    if static:
        from django_distill import distill_path  # pylint: disable=import-outside-toplevel
        universal_path = distill_path
    else:
        paths.append(path('admin/', admin_urls()))
    if is_dev():
        paths.append(path('error/', error_urls()))
    if option_is_set('toolbar'):  # pragma: no cover, tested in subprocess
        paths.append(universal_path('__debug__/', debug_toolbar_urls()))
    if sitemaps:
        extra = {'sitemaps': sitemaps}
        paths.append(universal_path('sitemap.xml', public(sitemap), extra, name='sitemap'))
    paths.append(universal_path('robots.txt', public(rules_list), name='robots'))
    return paths



from typing import Any, Callable, Dict, List, Optional, Sequence, Tuple, Union

from django.http import HttpResponseBase
from django.urls import URLPattern, URLResolver, include, path
from django.views.generic import TemplateView
from django_stubs_ext import StrOrPromise

from django_logikal.sitemap import StaticSitemap
from django_logikal.views import public as mark_public


class Template:
    def __init__(
        self,
        app: Optional[str] = None,
        public: bool = False,
        extra_context: Optional[Dict[str, Any]] = None,
    ):
        self.app = app
        self.public = public
        self.extra_context = extra_context
        self._path_priority: Dict[str, str] = {}

    def _view(
        self,
        name: str,
        public: Optional[bool] = None,
        extra_context: Optional[Dict[str, Any]] = None,
    ) -> Callable[..., HttpResponseBase]:
        extra_context = {**(self.extra_context or {}), **(extra_context or {})}
        template_name = f'{self.app + "/" if self.app else ""}{name}.html.j'
        view = TemplateView.as_view(template_name=template_name, extra_context=extra_context)
        public = public if public is not None else self.public
        return mark_public(view) if public else view

    def _add_path_priority(self, name: str, priority: Optional[str]) -> None:
        if priority:
            self._path_priority[name if not self.app else f'{self.app}:{name}'] = priority

    def path(  # pylint:disable=too-many-arguments
        self,
        route: StrOrPromise,
        name: str,
        public: Optional[bool] = None,
        priority: Optional[str] = None,
        extra_context: Optional[Dict[str, Any]] = None,
        **kwargs: Any,
    ) -> URLPattern:
        self._add_path_priority(name, priority)
        view = self._view(name, extra_context=extra_context, public=public)
        return path(route=route, view=view, name=name, **kwargs)

    def static_path(
        self,
        route: StrOrPromise,
        name: str,
        priority: Optional[str] = None,
        **kwargs: Any,
    ) -> Any:
        self._add_path_priority(name, priority)
        from django_distill import distill_path  # pylint: disable=import-outside-toplevel
        return distill_path(route=route, view=self._view(name), name=name, **kwargs)

    def include(
        self, paths: List[URLPattern],
    ) -> Tuple[Sequence[Union[URLResolver, URLPattern]], Optional[str], Optional[str]]:
        if not self.app:
            raise RuntimeError('The app name must be specified')
        return include((paths, self.app))

    def sitemap(self) -> StaticSitemap:
        return StaticSitemap(path_priority=self._path_priority)



from typing import Mapping, Sequence

from django.contrib.sitemaps import Sitemap
from django.urls import reverse


class StaticSitemap(Sitemap[str]):
    i18n = True
    alternates = True
    x_default = True

    def __init__(self, path_priority: Mapping[str, str]):
        self.path_priority = path_priority

    def items(self) -> Sequence[str]:
        return list(self.path_priority.keys())

    def priority(self, item: str) -> Sequence[str]:
        return self.path_priority[item]

    @staticmethod
    def location(item: str) -> str:
        return reverse(item)



import inspect
from typing import Any, Dict, Type

from django_logikal.env import get_option


class SettingsUpdateMeta(type):
    def __str__(cls) -> str:
        return f'{cls.__module__}.{cls.__qualname__}'


class SettingsUpdate(metaclass=SettingsUpdateMeta):
    @staticmethod
    def apply(settings: 'Settings') -> None:
        """
        Apply the current settings update to the provided settings.
        """


class Settings:
    def __init__(self, settings: Dict[str, Any]):
        """
        Store and update the provided settings.
        """
        self._settings = settings

    def __setitem__(self, key: str, value: Any) -> None:
        self._settings[key] = value

    def __getitem__(self, key: str) -> Any:
        return self._settings[key]

    def __contains__(self, item: str) -> bool:
        return item in self._settings

    def update(self, settings_update: Type[SettingsUpdate]) -> 'Settings':
        """
        Apply the given settings update to the stored settings.
        """
        # Apply updates from class attributes
        if get_option('log_level') == 'DEBUG':
            print(f'Settings update: {settings_update}')
        for attribute in dir(settings_update):
            if attribute == attribute.upper():
                self[attribute] = getattr(settings_update, attribute)

        # Apply dynamic updates from the entire class inheritance tree (starting at the root)
        for base in reversed(inspect.getmro(settings_update)):
            # Note: we only call apply when it is not inherited
            # (because we will walk up the inheritance tree anyways)
            if issubclass(base, SettingsUpdate) and 'apply' in base.__dict__:
                if get_option('log_level') == 'DEBUG':
                    print(f'Applying {base}')
                base.apply(self)

        return self



from pathlib import Path
from typing import Any

from django.contrib.staticfiles import storage


class ManifestStaticFilesStorage(storage.ManifestStaticFilesStorage):
    """
    Remove the unprocessed source files after post-processing.
    """
    default_template = 'url(\'%(url)s\')'  # patch the rewriter to use single quotes

    def post_process(self, *args: Any, **kwargs: Any) -> Any:
        for name, hashed_name, processed in super().post_process(*args, **kwargs):
            yield name, hashed_name, processed
            if processed:
                Path(self.path(name)).unlink()



from typing import Any, Callable, Dict, List

from django.contrib.auth.decorators import login_required
from django.http import HttpRequest

from django_logikal.middleware import Middleware
from django_logikal.views import PublicViewMixin


class LoginRequiredByDefaultMiddleware(Middleware):
    """
    Require login on all views by default.

    Public views must be marked explicitly via the :func:`~django_logikal.views.public` decorator
    or by inheriting from :class:`~django_logikal.views.PublicView` or
    :class:`~django_logikal.views.PublicViewMixin`.

    .. warning:: When this middleware is used, you must mark all public views explicitly, including
        views included from applications like :mod:`django.contrib.admin`. You may use the
        :ref:`URL utility functions <urls:URLs & Paths>` which provide such explicitly marked
        public views where necessary.

    .. note:: All :ref:`standard settings modules <settings:Settings>` include this middleware by
        default.
    """
    @staticmethod
    def process_view(
        request: HttpRequest,
        view_func: Callable[..., Any],
        view_args: List[Any],
        view_kwargs: Dict[str, Any],
    ) -> Any:
        if getattr(view_func, 'public_view', False):
            return None
        if issubclass(getattr(view_func, 'view_class', type(None)), PublicViewMixin):
            return None
        return login_required(view_func)(request, *view_args, **view_kwargs)



import os
import re
from datetime import datetime, tzinfo as tzinfo_class
from pathlib import Path
from typing import Any, Mapping, Optional

from babel import Locale
from babel.support import Format
from django.contrib.staticfiles import finders
from django.http.request import HttpRequest
from django.templatetags.static import static as django_static
from django.urls import reverse
from django.utils import timezone
from django.utils.safestring import SafeString, mark_safe
from django.utils.translation import get_language
from jinja2.runtime import Context
from jinja2.utils import pass_context

import django_logikal  # for type checking


@pass_context
def context(context: Context) -> Context:  # pylint: disable=redefined-outer-name
    # noqa: D400, D402, D415
    """
    context()

    Return the current template context.
    """
    return context


def static(path: str) -> str:
    """
    Return the absolute server URL path to a static asset.
    """
    return django_static(path)


def static_path(path: str) -> Path:
    """
    Return the local path to a static asset.
    """
    if not (file_path := finders.find(path)):
        raise RuntimeError(f'Static file "{path}" not found')
    return Path(file_path)


def include_static(path: str) -> SafeString:
    """
    Insert the contents of the given static file into the template.

    Useful for inlining CSS, SVG or JavaScript content.

    .. DANGER:: **Security risk: the contents of the referenced file are inserted unescaped.**

        Do not use this function to include non-trusted, user-generated or user-uploaded content,
        or content that can be in any way influenced by users.

    """
    file_path = static_path(path)
    content = file_path.read_text(encoding='utf-8').lstrip()
    if file_path.suffix == '.svg':
        content = re.sub(r'^<\?xml [^>]*\?>', '', content).lstrip()
    return mark_safe(content)  # nosec: danger is documented


def url(
    *args: Any,
    request: Optional[HttpRequest] = None,
    request_get_update: Optional[Mapping[str, Any]] = None,
    **kwargs: Any,
) -> str:
    """
    Return the absolute server URL path associated with the given view name.

    HTTP GET parameters are automatically appended when the request is provided.
    """
    path = reverse(*args, **kwargs)
    if request:
        request_get = request.GET.copy()
        request_get.update(request_get_update or {})
        if request_get:
            path = '?'.join([path, request_get.urlencode(safe='/')])
    return path


def url_name(request: HttpRequest) -> str:
    """
    Return the view name associated with the current request.
    """
    if not (url_match := request.resolver_match):
        raise RuntimeError('URL resolving has not taken place yet')
    return f'{url_match.app_name or ""}:{url_match.url_name or ""}'


def language() -> str:
    """
    Return the current language code.
    """
    return get_language()


def format(  # pylint: disable=redefined-builtin
    locale: Optional[Locale] = None,
    language_code: Optional[str] = None,
    tzinfo: Optional[tzinfo_class] = None,
) -> Format:
    """
    Return a locale-aware and time zone-aware formatter.

    Defaults to deriving the locale from the current language and using the current time zone.

    .. tip:: Often times you can use a single formatter instance in a template as follows:

        .. code-block:: jinja

            {% set fmt = format() %}

            ... {{ fmt.decimal(number) }} ...
            ... {{ fmt.datetime(timestamp) }} ...

        Note that the current locale can be influenced with the ``language`` tag, while the current
        time zone can be influenced with the ``timezone`` tag:

        .. code-block:: jinja

            {% language 'en-us' %}
            {% timezone 'Europe/London' %}

            {{ format().datetime(timestamp, format='long') }}
            <!-- if timestamp is datetime(2023, 7, 1, 14, 34, 56, tzinfo=timezone.utc) -->
            <!-- displays 1 July 2023, 15:34:56 BST -->

            {% endtimezone %}
            {% endlanguage %}

    """
    return Format(
        locale=locale or Locale.parse(language_code or get_language(), sep='-'),
        tzinfo=tzinfo or timezone.get_current_timezone(),
    )


def cwd() -> Path:
    """
    Return the current working directory.
    """
    return Path(os.getcwd())


def now() -> datetime:
    """
    Return the current date and time.
    """
    return timezone.now()


def bibliography(name: str) -> 'django_logikal.bibliography.Bibliography':
    """
    Return a :class:`~django_logikal.bibliography.Bibliography` instance.

    Args:
        name: The name of the bibliography configuration to use.

    .. note:: Requires the :ref:`bibliography extra <index:Bibliography>`.

    Usage
    -----
    First, make sure that your bibliographies are in the appropriate template folders and that
    their paths are added to the Django settings file as follows:

    .. code-block::

        BIBLIOGRAPHIES = {'references': 'website/references.bib'}

    Once configured, you can use this function in templates as follows:

    .. code-block:: jinja

        {% set bib = bibliography('references') %}
        {% set cite = bib.cite %}

        ... {{ cite('entry') }} ...

        {{ bib.references() }}

    """
    from django_logikal.bibliography import Bibliography  # pylint: disable=import-outside-toplevel

    return Bibliography(name=name)
